import numpy as np
import ujson


class ImageMapperArgs:
    def __init__(self, cam_u, cam_l, cam_f, cam_angles, camera_wh_px, cam_fu, cam_fl, cam_lu=0):
        """
        :type cam_u: float
        :type cam_l: float
        :type cam_f: float
        :type cam_angles:(float, float)
        :type camera_wh_px: (int, int)
        :type cam_fu: float
        :type cam_fl: float
        :type cam_lu: float
        """
        self.cam_f = cam_f    # forward
        self.cam_l = cam_l    # left
        self.cam_u = cam_u    # up

        self.cam_fu = cam_fu  # elevation
        self.cam_fl = cam_fl  # azimuth
        self.cam_lu = cam_lu  # roll

        self.cam_angles = cam_angles
        self.camera_wh_px = camera_wh_px

    @staticmethod
    def deserialize(jnode):
        """
        :type jnode: dict
        :rtype: ImageMapperArgs
        :raise ValueError:
        """
        # a = {
        #     'sensor': {
        #         'dimension': {
        #             'fov_h': 14.2,
        #             'fov_v': 8.0,
        #             'width': 1920,
        #             'height': 1200
        #         },
        #         'position': {
        #             'forward': 0,
        #             'left': 0,
        #             'up': 475
        #         },
        #         'rotation': {
        #             'elevation': 10,
        #             'azimuth': 15,
        #             'pitch': 0
        #         }
        #     }
        # }

        sensor_jnode = jnode.get('sensor')
        if sensor_jnode is None:
            raise ValueError

        dimension_jnode = sensor_jnode.get('dimension')
        position_jnode = sensor_jnode.get('position')
        rotation_jnode = sensor_jnode.get('rotation')

        if None in [dimension_jnode, position_jnode, rotation_jnode]:
            raise ValueError

        cam_f = position_jnode.get('forward', 0)
        cam_l = position_jnode.get('left', 0)
        cam_u = position_jnode.get('up', 0)

        cam_fu = rotation_jnode['elevation']
        cam_fl = rotation_jnode['azimuth']
        cam_lu = rotation_jnode.get('pitch', 0)

        cam_fov_hv = dimension_jnode['fov_h'], dimension_jnode['fov_v']
        cam_image_wh = dimension_jnode['width'], dimension_jnode['height']

        return ImageMapperArgs(
            cam_u=cam_u,
            cam_l=cam_l,
            cam_f=cam_f,
            cam_fu=cam_fu,
            cam_fl=cam_fl,
            cam_lu=cam_lu,
            cam_angles=cam_fov_hv,
            camera_wh_px=cam_image_wh)

    @staticmethod
    def from_file(filepath):
        try:
            with open(filepath) as conf_jnode:
                conf_jnode = ujson.load(conf_jnode)

        except Exception as x:
            raise IOError('failed to load %s (%s)' % (filepath, x))

        return ImageMapperArgs.deserialize(conf_jnode)


class Fov:
    def __init__(self, raw_matrix_4x4, inv_matrix_3x3, vector_1x3, cam_wh):
        self._image_mat_4x4 = raw_matrix_4x4
        self._inv_matrix_3x3 = inv_matrix_3x3
        self._vector_1x3 = vector_1x3
        self._image_wh = cam_wh

    @staticmethod
    def create(cam_u, cam_l, cam_f, cam_fov_hv, cam_wh, cam_fu, cam_fl, cam_lu=0):
        """
        :type cam_u: float
        :type cam_l: float
        :type cam_f: float
        :type cam_fov_hv:(float, float)
        :type cam_wh: (int, int)
        :type cam_fu: float
        :type cam_fl: float
        :type cam_lu: float
        :rtype: Fov
        """
        # define intristic camera params
        def mk_intristic_mat():
            half_w, half_h = 0.5 * cam_wh[0], 0.5 * cam_wh[1]
            tan_half_cam_fov_h = np.tan(np.deg2rad(0.5 * cam_fov_hv[0]))
            tan_half_cam_fov_v = np.tan(np.deg2rad(0.5 * cam_fov_hv[1]))

            return np.matrix([
                [half_w / tan_half_cam_fov_h,                           0, half_w, 0],
                [0,                           half_h / tan_half_cam_fov_v, half_h, 0],
                [0,                                                     0,      1, 0],
                [0,                                                     0,      0, 0]
            ])

        def mk_view_mat():
            rad_cam_fu = np.deg2rad(cam_fu)  # elevation
            rad_cam_fl = np.deg2rad(cam_fl)  # azimuth
            rad_cam_lu = np.deg2rad(cam_lu)  # roll

            # elevation matrix
            fu_mat = np.matrix([
                [1,                   0,                   0, 0],
                [0, +np.cos(rad_cam_fu), -np.sin(rad_cam_fu), 0],
                [0, +np.sin(rad_cam_fu), +np.cos(rad_cam_fu), 0],
                [0,                   0,                   0, 1]
            ])

            # azimuth matrix
            fl_mat = np.matrix([
                [+np.cos(rad_cam_fl), 0, +np.sin(rad_cam_fl), 0],
                [0,                   1,                   0, 0],
                [-np.sin(rad_cam_fl), 0, +np.cos(rad_cam_fl), 0],
                [0,                   0,                   0, 1]
            ])

            # roll matrix
            lu_mat = np.matrix([
                [+np.cos(rad_cam_lu), -np.sin(rad_cam_lu), 0, 0],
                [+np.sin(rad_cam_lu), +np.cos(rad_cam_lu), 0, 0],
                [0,                                     0, 1, 0],
                [0, 0, 0, 1]
            ])

            # position matrix
            pos_mat = np.matrix([
                [1, 0, 0, -cam_l],
                [0, 1, 0, -cam_u],
                [0, 0, 1, -cam_f],
                [0, 0, 0,      1]
            ])

            return fl_mat * lu_mat * fu_mat * pos_mat

        image_mat = mk_intristic_mat() * mk_view_mat()

        # get "top-front" 3x3 matrix fom big one
        m_mat = np.matrix(image_mat[0:3, 0:3])
        inv_m_mat = m_mat.getI()

        # get "top-rear" 1x3 matrix
        p4 = np.matrix(image_mat[0:3, 3])

        c_tilde = - inv_m_mat * p4

        return Fov(image_mat, inv_m_mat, c_tilde, cam_wh)

    @staticmethod
    def create2(args):
        """
        :type args: ImageMapperArgs
        :rtype: Fov
        """
        return Fov.create(cam_u=args.cam_u,
                          cam_l=args.cam_l,
                          cam_f=args.cam_f,
                          cam_fov_hv=args.cam_angles,
                          cam_wh=args.camera_wh_px,
                          cam_fu=args.cam_fu,
                          cam_fl=args.cam_fl,
                          cam_lu=args.cam_lu)

    def mk_world_point(self, image_x, image_y, world_t):
        """
        :type image_x: int
        :type image_y: int
        :type world_t: float
        :rtype: (float, float)
        """
        image_xyw = np.matrix([[image_x], [image_y], [1]])

        world_candidate = self._inv_matrix_3x3 * image_xyw

        # transform to world point where = world_t
        podgonian = -(self._vector_1x3[1]) / (world_candidate[1]) + world_t / (world_candidate[1])

        world_point = podgonian[0, 0] * self._inv_matrix_3x3 * image_xyw + self._vector_1x3

        world_f = world_point[2, 0]
        world_r = world_point[0, 0]
        return world_f, world_r

    def mk_image_point(self, world_f, world_r, world_t):
        """
        :type world_r: float
        :type world_t: float
        :type world_f: float
        :rtype: (float, float)
        """
        world_rtfw = np.matrix([[world_r], [world_t], [world_f], [1]])
        image_homo_point = self._image_mat_4x4 * world_rtfw

        # normalize
        image_point = image_homo_point / image_homo_point[2]  # TODO
        image_x, image_y, _u, _i = image_point

        # for y we revert result
        rel_image_x = image_x / float(self._image_wh[0])
        rel_image_y = 1.0 - image_y / float(self._image_wh[1])
        return rel_image_x, rel_image_y

    def get_dist_image(self, rel_image_x, rel_image_y, world_t):
        """
        :param rel_image_x: visible values are in range [0.0; 1.0]
        :type rel_image_x: float

        :param rel_image_y: visible values are in range [0.0; 1.0]
        :type rel_image_y: float

        :param world_t: meters
        :type world_t: float

        :rtype: float
        """
        image_x = int(rel_image_x * self._image_wh[0])
        image_y = int((1.0 - rel_image_y) * self._image_wh[1])

        world_f, world_r = Fov.mk_world_point(self, image_x, image_y, world_t)
        return world_f
