from itertools import ifilter

from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator


# noinspection PyUnusedLocal
def im_get_creators(entry_uimod, all_used_uimods, uimod_to_newconf_map, forced=False):
    """
    :type entry_uimod: ph.UiMod | ph.UiMod2
    :type all_used_uimods: set[ph.UiMod, ph.UiMod2]
    :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
    :type uimod_to_newconf_map: dict
    :type forced: bool
    :rtype: list of BaseCreator
    """
    if entry_uimod.type != ph.UI_TYPE_LOCATION:
        return []

    try:
        cam_uimod = next(ifilter(lambda x: x.type == ph.UI_TYPE_MV_CAMERA, all_used_uimods))
        location_uimod = next(ifilter(lambda x: x.type == ph.UI_TYPE_LOCATION, all_used_uimods))

        activate = any([
            uimod_to_newconf_map.get(location_uimod),
            uimod_to_newconf_map.get(cam_uimod)
        ])

        if forced or activate:
            return [ImCreator(location_uimod, cam_uimod)]

    except StopIteration:
        pass

    return []


class ImCreator(BaseCreator):
    def __init__(self, location_uimod, cam_uimod):
        """
        :type location_uimod: ph.UiMod
        :type cam_uimod: ph.UiMod
        """
        BaseCreator.__init__(self)
        self._location_uimod = location_uimod
        self._cam_uimod = cam_uimod
        self._im_dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_IMAGE_MAPPER, location_uimod.pid, location_uimod.id)
        self._info_main = None

    @property
    def dgmod(self):
        """
        :rtype: dg.DgMod
        """
        return self._im_dgmod

    def prepare(self, uimod_to_newconf_map):
        """
        :param uimod_to_newconf_map: Dict[UiMod | UiMod2, Jnode]
        :type uimod_to_newconf_map: dict
        """
        location_uiconf = uimod_to_newconf_map.get(self._location_uimod) or self._location_uimod.load_conf()
        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()

        camera_jnode = location_uiconf['rw']['args']['camera']

        camera_position_jnode = camera_jnode['position']
        forward = camera_position_jnode['forward']
        left = camera_position_jnode['left']
        up = camera_position_jnode['up']

        # deserialize sensor-rotation
        camera_rotation_manual_jnode = camera_jnode['rotation']
        azimuth = camera_rotation_manual_jnode['azimuth']
        elevation = camera_rotation_manual_jnode['elevation']
        roll = camera_rotation_manual_jnode['roll']

        fov_h, fov_v = cam_uiconf['ro']['fov-hv']

        ui_acquisition_jnode = cam_uiconf['rw']['videostream']
        ui_sensor_jnode = ui_acquisition_jnode['resolution']

        # create main.conf
        main_conf = {
            "sensor": {
                "dimension": {
                    "fov_h": fov_h,
                    "fov_v": fov_v,
                    "width": ui_sensor_jnode['w'],
                    "height": ui_sensor_jnode['h']
                },
                "position": {
                    "forward": forward,
                    "left": left,
                    "up": up
                },
                "rotation": {
                    "elevation": elevation,
                    "azimuth": azimuth,
                    "pitch": roll
                }
            }
        }

        self._info_main = self._im_dgmod, main_conf

    def complete(self):
        edge_im, main_conf = self._info_main  # type: dg.DgMod, dict
        edge_im.install()
        edge_im.save_main_conf(main_conf)

    def rollback(self):
        pass

    def notify(self):
        """
        :rtype: None
        """
        pass


class ImDeleter:
    def __init__(self, dgmod):
        """
        :type dgmod: dg.DgMod
        """
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod):
        """
        :type entry_uimod: ph.UiMod
        :rtype: ImDeleter | None
        """
        if entry_uimod.type == ph.UI_TYPE_LOCATION:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_IMAGE_MAPPER, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self):
        self._dgmod.uninstall()
